/*
 * Copyright (c) 2003, 2004 Nokia
 * Author: Timo Savola <tsavola@movial.fi>
 *
 * This program is licensed under GPL (see COPYING for details)
 *
 * $Id: misc_runner.c,v 1.10 2004/02/11 16:47:53 tsavola Exp $
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <time.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/types.h>

#define SBRSH_CMD   "/scratchbox/tools/bin/sbrsh"
#define CONFIG_NAME "/targets/links/scratchbox.config"
#define LOGNAME_FMT "/tmp/cputransp_%s.log"

#define ERROR_CODE  126

static const char *progname;

static void error(const char *s)
{
	if (s) {
		fprintf(stderr, "%s: %s: %s\n", progname, s, strerror(errno));
	} else {
		fprintf(stderr, "%s: %s\n", progname, strerror(errno));
	}
	exit(ERROR_CODE);
}

static int read_line(FILE *file,
		     char *buf,
		     int size)
{
	int comm = 0;
	int c, i = 0;

	if (feof(file)) {
		return -1;
	}

	while (1) {
		c = fgetc(file);
		if (c == EOF || c == '\n') {
			break;
		}
		if (c == '#') {
			comm = 1;
		}
		if (!comm && i < size - 1) {
			*buf++ = c;
			i++;
		}
	}
	*buf = '\0';

	return i;
}

static int is_set(char *var)
{
	char *str = getenv(var);
	return (str && strcmp(str, "no") != 0);
}

static void read_config(void)
{
	FILE *file;
	char str[1024];

	file = fopen(CONFIG_NAME, "r");
	if (!file) {
		error(CONFIG_NAME);
	}

	while (read_line(file, str, sizeof (str)) >= 0) {
		if (strchr(str, '=') != NULL) {
			putenv(strdup(str));
		}
	}

	fclose(file);
}

static char *build_log_name(void)
{
	char *user, *str;
	int len;

	user = getenv("LOGNAME");
	if (!user) {
		user = getenv("USER");
		if (!user) {
			fprintf(stderr, "%s: LOGNAME and USER not set\n",
				progname);
			exit(ERROR_CODE);
		}
	}

	len = strlen(LOGNAME_FMT) - 2 + strlen(user) + 1;

	str = malloc(len);
	if (!str) {
		error(NULL);
	}

	snprintf(str, len, LOGNAME_FMT, user);

	return str;
}

static void print_prefix(FILE *file,
			 struct timeval *tv)
{
	struct tm *t;

	t = gmtime(&tv->tv_sec);

	fprintf(file, "[%02d.%02d.%04d %02d:%02d:%02d %5d] ",
		t->tm_mday, t->tm_mon + 1, 1900 + t->tm_year,
		t->tm_hour, t->tm_min, t->tm_sec, getpid());
}

static void print_cmd(FILE *file,
		      const char *pwd,
		      char **args)
{
	int i;

	fprintf(file, "pwd: \"%s\"", pwd);

	fprintf(file, "  cmd: \"");
	for (i = 0; args[i]; i++) {
		fprintf(file, i ? " %s" : "%s", args[i]);
	}
	fprintf(file, "\"\n");

	fflush(file);
}

static void print_time(FILE *file,
		       struct timeval *start,
		       struct timeval *end)
{
	int sec, micro;

	sec = end->tv_sec - start->tv_sec;
	micro = end->tv_usec - start->tv_usec;

	if (micro < 0) {
		sec--;
		micro += 1000000;
	} else if (micro >= 1000000) {
		sec++;
		micro -= 1000000;
	}

	fprintf(file, "time: %d.%06d\n", sec, micro);
	fflush(file);
}

static void fix_args(int argc,
		     char **argv)
{
	FILE *file;
	char buf[PATH_MAX];

	if ((argc < 2 || strcmp(argv[0], argv[1]) != 0) &&
	    (argc < 3 || strcmp(argv[0], argv[2]) != 0)) {
		return;
	}

	file = fopen(argv[0], "r");
	if (!file) {
		error(argv[0]);
	}

	if (fread(buf, 1, 2, file) != 2) {
		error(argv[0]);
	}

	if (strncmp(buf, "#!", 2) == 0) {
		if (read_line(file, buf, sizeof (buf)) > 0) {
			char *cmd;
			int end;

			for (cmd = buf; isspace(*cmd); cmd++);
			for (end = 0; cmd[end] && !isspace(cmd[end]); end++);

			if (end > 0) {
				argv[0] = strndup(cmd, end);
			}
		}
	}

	fclose(file);
}

static char **build_args(char *pwd,
			 int argc,
			 char **argv)
{
	char *target, *filter, **vec;
	int argstart, len, i;

	target = getenv("SBOX_TARGET_NAME");
	if (!target || strlen(target) == 0) {
		fprintf(stderr, "%s: SBOX_TARGET_NAME not set\n", progname);
		exit(ERROR_CODE);
	}

	filter = getenv("SBOX_CPUTRANSPARENCY_FILTER");
	if (filter && '\0' == filter[0]) {
		filter = NULL;
	}

	argstart = 4;

	if (filter) {
		argstart++;
	}

	len = argstart + (argc - 1);
	vec = calloc(len + 1, sizeof (char *));
	if (!vec) {
		error(NULL);
	}

	vec[0] = SBRSH_CMD;
	vec[1] = target;
	vec[2] = "-d";
	vec[3] = pwd;

	if (filter) {
		vec[4] = filter;
	}

	for (i = 1; i < argc; i++) {
		vec[argstart + (i - 1)] = argv[i];
	}

	fix_args(len - argstart, &vec[argstart]);

	return vec;
}

static void dummy_handler(int sig)
{
}

int main(int argc,
	 char **argv)
{
	char *logname, *pwd, **argvec;
	FILE *logfile;
	struct timeval start_time;
	struct sigaction action;
	pid_t pid;

	progname = argv[0];

	if (argc < 2) {
		return 0;
	}

	if (is_set("SBOX_DISABLE_CPUTRANSPARENCY")) {
		fprintf(stderr, "%s: cannot execute binary file\n", argv[1]);
		exit(ERROR_CODE);
	}

	read_config();

	logname = build_log_name();

	if (is_set("SBOX_CPUTRANSPARENCY_VERBOSE")) {
		fprintf(stderr, "\n ### INFO ###  cpuTransparency!"
			"  See %s for details!\n", logname);
	}

	pwd = get_current_dir_name();
	if (!pwd) {
		error(NULL);
	}

	argvec = build_args(pwd, argc, argv);

	logfile = fopen(logname, "a");
	if (!logfile) {
		error(logname);
	}

	gettimeofday(&start_time, NULL);
	print_prefix(logfile, &start_time);
	print_cmd(logfile, pwd, &argvec[4]);

	action.sa_handler = dummy_handler;
	sigemptyset(&action.sa_mask);
	action.sa_flags = 0;
	sigaction(SIGCHLD, &action, NULL);

	pid = fork();
	if (pid < 0) {
		error(NULL);
	}

	if (pid == 0) {
		execvp(argvec[0], argvec);
		error(argvec[0]);
	}

	while (1) {
		int status, rc;
		struct timeval end_time;

		if (waitpid(pid, &status, 0) < 0) {
			if (errno == EINTR) {
				continue;
			}

			if (errno != ECHILD) {
				int e = errno;
				kill(pid, SIGTERM);
				errno = e;
			}

			error(NULL);
		}

		gettimeofday(&end_time, NULL);
		print_prefix(logfile, &end_time);

		rc = ERROR_CODE;

		if (WIFEXITED(status)) {
			rc = WEXITSTATUS(status);
			fprintf(logfile, "rc: %d  ", rc);
		}

		if (WIFSIGNALED(status)) {
			fprintf(logfile, "signal: \"%s\"  ",
				strsignal(WTERMSIG(status)));
		}

		print_time(logfile, &start_time, &end_time);

		return rc;
	}

	return 0;
}
